
import logging
import multiprocessing
import queue
import time

from gi.repository import GLib

from meld.matchers import myers

log = logging.getLogger(__name__)


class MatcherWorker(multiprocessing.Process):

    END_TASK = -1

    matcher_class = myers.InlineMyersSequenceMatcher

    def __init__(self, tasks, results):
        super().__init__()
        self.tasks = tasks
        self.results = results
        self.daemon = True

    def run(self):
        while True:
            task_id, (text1, textn) = self.tasks.get()
            if task_id == self.END_TASK:
                break

            try:
                matcher = self.matcher_class(None, text1, textn)
                self.results.put((task_id, matcher.get_opcodes()))
            except Exception as e:
                log.error("Exception while running diff: %s", e)
            time.sleep(0)


class CachedSequenceMatcher:
    """Simple class for caching diff results, with LRU-based eviction

    Results from the SequenceMatcher are cached and timestamped, and
    subsequently evicted based on least-recent generation/usage. The LRU-based
    eviction is overly simplistic, but is okay for our usage pattern.
    """

    TASK_GRACE_PERIOD = 1

    def __init__(self, scheduler):
        """Create a new caching sequence matcher

        :param scheduler: a `meld.task.SchedulerBase` used to schedule
            sequence comparison result checks
        """
        self.scheduler = scheduler
        self.cache = {}
        self.tasks = multiprocessing.Queue()
        self.tasks.cancel_join_thread()
        # Limiting the result queue here has the effect of giving us
        # much better interactivity. Without this limit, the
        # result-checker tends to get starved and all highlights get
        # delayed until we're almost completely finished.
        self.results = multiprocessing.Queue(5)
        self.results.cancel_join_thread()
        self.thread = MatcherWorker(self.tasks, self.results)
        self.task_id = 1
        self.queued_matches = {}
        GLib.idle_add(self.thread.start)

    def stop(self) -> None:
        self.tasks.put((MatcherWorker.END_TASK, ('', '')))
        if self.thread.is_alive():
            self.thread.join(self.TASK_GRACE_PERIOD)
            if self.thread.exitcode is None:
                self.thread.terminate()
        self.cache = {}
        self.queued_matches = {}

    def match(self, text1, textn, cb):
        texts = (text1, textn)
        try:
            self.cache[texts][1] = time.time()
            opcodes = self.cache[texts][0]
            GLib.idle_add(lambda: cb(opcodes))
        except KeyError:
            GLib.idle_add(lambda: self.enqueue_task(texts, cb))

    def enqueue_task(self, texts, cb):
        if not bool(self.queued_matches):
            self.scheduler.add_task(self.check_results)
        self.queued_matches[self.task_id] = (texts, cb)
        self.tasks.put((self.task_id, texts))
        self.task_id += 1

    def check_results(self):
        try:
            task_id, opcodes = self.results.get(block=True, timeout=0.01)
            texts, cb = self.queued_matches.pop(task_id)
            self.cache[texts] = [opcodes, time.time()]
            GLib.idle_add(lambda: cb(opcodes))
        except queue.Empty:
            pass

        return bool(self.queued_matches)

    def clean(self, size_hint):
        """Clean the cache if necessary

        @param size_hint: the recommended minimum number of cache entries
        """
        if len(self.cache) <= size_hint * 3:
            return
        items = list(self.cache.items())
        items.sort(key=lambda it: it[1][1])
        for item in items[:-size_hint * 2]:
            del self.cache[item[0]]



# Copyright (C) 2016 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gdk, Gio, GObject

from meld.settings import load_settings_schema

WINDOW_STATE_SCHEMA = 'org.gnome.meld.WindowState'


class SavedWindowState(GObject.GObject):
    '''Utility class for saving and restoring GtkWindow state'''

    __gtype_name__ = 'SavedWindowState'

    width = GObject.Property(
        type=int, nick='Current window width', default=-1)
    height = GObject.Property(
        type=int, nick='Current window height', default=-1)
    is_maximized = GObject.Property(
        type=bool, nick='Is window maximized', default=False)
    is_fullscreen = GObject.Property(
        type=bool, nick='Is window fullscreen', default=False)

    def bind(self, window):
        window.connect('size-allocate', self.on_size_allocate)
        window.connect('window-state-event', self.on_window_state_event)

        # Don't re-read from gsettings after initialisation; we've seen
        # what looked like issues with buggy debounce here.
        bind_flags = (
            Gio.SettingsBindFlags.DEFAULT |
            Gio.SettingsBindFlags.GET_NO_CHANGES
        )
        self.settings = load_settings_schema(WINDOW_STATE_SCHEMA)
        self.settings.bind('width', self, 'width', bind_flags)
        self.settings.bind('height', self, 'height', bind_flags)
        self.settings.bind('is-maximized', self, 'is-maximized', bind_flags)
        self.settings.bind('is-fullscreen', self, 'is-fullscreen', bind_flags)

        window.set_default_size(self.props.width, self.props.height)
        if self.props.is_maximized:
            window.maximize()
        if self.props.is_fullscreen:
            window.fullscreen()

    def on_size_allocate(self, window, allocation):
        if not (self.props.is_maximized or self.props.is_fullscreen):
            width, height = window.get_size()
            if width != self.props.width:
                self.props.width = width
            if height != self.props.height:
                self.props.height = height

    def on_window_state_event(self, window, event):
        state = event.window.get_state()

        is_maximized = state & Gdk.WindowState.MAXIMIZED
        if is_maximized != self.props.is_maximized:
            self.props.is_maximized = is_maximized

        is_fullscreen = state & Gdk.WindowState.FULLSCREEN
        if is_fullscreen != self.props.is_fullscreen:
            self.props.is_fullscreen = is_fullscreen



# Copyright (C) 2010-2015 Kai Willadsen <kai.willadsen@gmail.com>
# Copyright (C)      2016 Guillaume Hoffmann <guillaumh@gmail.com>
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import errno
import os
import shutil
import subprocess
import tempfile
from collections import defaultdict

from . import _vc


class Vc(_vc.Vc):

    # Requires Darcs version >= 2.10.3
    # TODO implement get_commits_to_push_summary using `darcs push --dry-run`
    # Currently `darcs whatsnew` (as of v2.10.3) does not report conflicts
    # see http://bugs.darcs.net/issue2138

    CMD = "darcs"
    NAME = "Darcs"
    VC_DIR = "_darcs"

    state_map = {
        "a": _vc.STATE_NONE,
        "A": _vc.STATE_NEW,
        "M": _vc.STATE_MODIFIED,
        "M!": _vc.STATE_CONFLICT,
        "R": _vc.STATE_REMOVED,
        "F": _vc.STATE_NONEXIST,  # previous name of file
        "T": _vc.STATE_RENAMED,   # new name of file
    }

    @classmethod
    def is_installed(cls):
        try:
            proc = _vc.popen([cls.CMD, '--version'])
            # check that version >= 2.10.3
            (x, y, z) = proc.read().split(" ", 1)[0].split(".", 2)[:3]
            assert (x, y, z) >= (2, 10, 3)
            return True
        except Exception:
            return False

    def commit(self, runner, files, message):
        command = [self.CMD, 'record', '-a', '-m', message]
        runner(command, [], refresh=True, working_dir=self.root)

    def update(self, runner):
        command = [self.CMD, 'pull', '-a']
        runner(command, [], refresh=True, working_dir=self.root)

    def push(self, runner):
        command = [self.CMD, 'push', '-a']
        runner(command, [], refresh=True, working_dir=self.root)

    def add(self, runner, files):
        command = [self.CMD, 'add', '-r']
        runner(command, files, refresh=True, working_dir=self.root)

    def remove(self, runner, files):
        command = [self.CMD, 'remove', '-r']
        runner(command, files, refresh=True, working_dir=self.root)

    def revert(self, runner, files):
        command = [self.CMD, 'revert', '-a']
        runner(command, files, refresh=True, working_dir=self.root)

    def get_path_for_repo_file(self, path, commit=None):
        if commit is not None:
            raise NotImplementedError()

        if not path.startswith(self.root + os.path.sep):
            raise _vc.InvalidVCPath(self, path, "Path not in repository")

        # `darcs show contents` needs the path before rename
        if path in self._reverse_rename_cache:
            path = self._reverse_rename_cache[path]

        path = path[len(self.root) + 1:]
        suffix = os.path.splitext(path)[1]
        process = subprocess.Popen(
            [self.CMD, "show", "contents", path], cwd=self.root,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        with tempfile.NamedTemporaryFile(prefix='meld-tmp',
                                         suffix=suffix, delete=False) as f:
            shutil.copyfileobj(process.stdout, f)
        return f.name

    @classmethod
    def valid_repo(cls, path):
        return not _vc.call([cls.CMD, "show", "repo", "--no-files"], cwd=path)

    def _update_tree_state_cache(self, path):
        # FIXME: currently ignoring 'path' due to darcs's bad
        #        behaviour (= fails) when given "" argument
        """ Update the state of the file(s) at self._tree_cache['path'] """
        while 1:
            try:
                proc = _vc.popen(
                    [self.CMD, "whatsnew", "-sl", "--machine-readable"],
                    cwd=self.location)
                lines = proc.read().split("\n")[:-1]
                break
            except OSError as e:
                if e.errno != errno.EAGAIN:
                    raise

        # Files can appear twice in the list if were modified and renamed
        # at once. Darcs first show file moves then modifications.
        if len(lines) == 0 and os.path.isfile(path):
            # If we're just updating a single file there's a chance that it
            # was it was previously modified, and now has been edited so that
            # it is un-modified.  This will result in an empty 'entries' list,
            # and self._tree_cache['path'] will still contain stale data.
            # When this corner case occurs we force self._tree_cache['path']
            # to STATE_NORMAL.
            self._tree_cache[path] = _vc.STATE_NORMAL
        else:
            tree_cache = defaultdict(int)
            tree_meta_cache = defaultdict(list)
            self._rename_cache = rename_cache = {}
            self._reverse_rename_cache = {}
            old_name = None
            for line in lines:
                # skip empty lines and line starting with "What's new in foo"
                if (not line.strip()) or line.startswith("What"):
                    continue
                statekey, name = line.split(" ", 1)
                name = os.path.normpath(name)
                if statekey == "F":
                    old_name = name

                path = os.path.join(self.location, name)

                if statekey == "T" and old_name:
                    old_path = os.path.join(self.location, old_name)
                    rename_cache[old_path] = path
                    old_name = None

                state = self.state_map.get(statekey.strip(), _vc.STATE_NONE)
                tree_cache[path] = state

            for old, new in rename_cache.items():
                self._reverse_rename_cache[new] = old
                old_name = old[len(self.root) + 1:]
                new_name = new[len(self.root) + 1:]
                tree_meta_cache[new] = ("%s ➡ %s" % (old_name, new_name))

            self._tree_cache.update(
                dict((x, y) for x, y in tree_cache.items()))
            self._tree_meta_cache = dict(tree_meta_cache)



# Copyright (C) 2002-2009 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2008-2009, 2013, 2019 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GObject, Gtk


@Gtk.Template(resource_path='/org/gnome/meld/ui/notebook-label.ui')
class NotebookLabel(Gtk.EventBox):

    __gtype_name__ = 'NotebookLabel'

    label = Gtk.Template.Child()

    label_text = GObject.Property(
        type=str,
        nick='Text of this notebook label',
        default='',
    )

    page = GObject.Property(
        type=object,
        nick='Notebook page for which this is the label',
        default=None,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.bind_property(
            'label-text', self.label, 'label',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
        )
        self.bind_property(
            'label-text', self, 'tooltip-text',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
        )

    @Gtk.Template.Callback()
    def on_label_button_press_event(self, widget, event):
        # Middle-click on the tab closes the tab.
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 2:
            self.page.on_delete_event()

    @Gtk.Template.Callback()
    def on_close_button_clicked(self, widget):
        self.page.on_delete_event()



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010, 2012-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Tuple

import cairo
from gi.repository import Gdk, GdkPixbuf, GObject, Gtk


class EmblemCellRenderer(Gtk.CellRenderer):

    __gtype_name__ = "EmblemCellRenderer"

    icon_cache: Dict[Tuple[str, int], GdkPixbuf.Pixbuf] = {}

    icon_name = GObject.Property(
        type=str,
        nick='Named icon',
        blurb='Name for base icon',
        default='text-x-generic',
    )

    emblem_name = GObject.Property(
        type=str,
        nick='Named emblem icon',
        blurb='Name for emblem icon to overlay',
    )

    secondary_emblem_name = GObject.Property(
        type=str,
        nick='Named secondary emblem icon',
        blurb='Name for secondary emblem icon to overlay',
    )

    icon_tint = GObject.Property(
        type=Gdk.RGBA,
        nick='Icon tint',
        blurb='GDK-parseable color to be used to tint icon',
    )

    def __init__(self):
        super().__init__()
        self._state = None
        # FIXME: hardcoded sizes
        self._icon_size = 16
        self._emblem_size = 8

    def _get_pixbuf(self, name, size):
        if (name, size) not in self.icon_cache:
            icon_theme = Gtk.IconTheme.get_default()
            pixbuf = icon_theme.load_icon(name, size, 0).copy()
            self.icon_cache[(name, size)] = pixbuf
        return self.icon_cache[(name, size)]

    def do_render(self, context, widget, background_area, cell_area, flags):
        context.translate(cell_area.x, cell_area.y)
        context.rectangle(0, 0, cell_area.width, cell_area.height)
        context.clip()

        # TODO: Incorporate padding
        context.push_group()
        if self.icon_name:
            pixbuf = self._get_pixbuf(self.icon_name, self._icon_size)
            context.set_operator(cairo.OPERATOR_SOURCE)
            # Assumes square icons; may break if we don't get the requested
            # size
            height_offset = int((cell_area.height - pixbuf.get_height()) / 2)
            Gdk.cairo_set_source_pixbuf(context, pixbuf, 0, height_offset)
            context.rectangle(0, height_offset,
                              pixbuf.get_width(), pixbuf.get_height())
            context.fill()

            if self.icon_tint:
                c = self.icon_tint
                r, g, b = c.red, c.green, c.blue
                # Figure out the difference between our tint colour and an
                # empirically determined (i.e., guessed) satisfying luma and
                # adjust the base colours accordingly
                luma = (r + r + b + g + g + g) / 6.
                extra_luma = (1.2 - luma) / 3.
                r, g, b = [min(x + extra_luma, 1.) for x in (r, g, b)]
                context.set_source_rgba(r, g, b, 0.4)
                context.set_operator(cairo.OPERATOR_ATOP)
                context.paint()

            if self.emblem_name:
                pixbuf = self._get_pixbuf(self.emblem_name, self._emblem_size)
                x_offset = self._icon_size - self._emblem_size
                context.set_operator(cairo.OPERATOR_OVER)
                Gdk.cairo_set_source_pixbuf(context, pixbuf, x_offset, 0)
                context.rectangle(x_offset, 0,
                                  cell_area.width, self._emblem_size)
                context.fill()

            if self.secondary_emblem_name:
                pixbuf = self._get_pixbuf(
                    self.secondary_emblem_name, self._emblem_size)
                x_offset = self._icon_size - self._emblem_size
                y_offset = self._icon_size - self._emblem_size + height_offset
                context.set_operator(cairo.OPERATOR_OVER)
                Gdk.cairo_set_source_pixbuf(
                    context, pixbuf, x_offset, y_offset)
                context.rectangle(
                    x_offset, y_offset, cell_area.width, self._emblem_size)
                context.fill()

        context.pop_group_to_source()
        context.set_operator(cairo.OPERATOR_OVER)
        context.paint()

    def do_get_size(self, widget, cell_area):
        # TODO: Account for cell_area if we have alignment set
        x_offset, y_offset = 0, 0
        width, height = self._icon_size, self._icon_size
        # TODO: Account for padding
        return (x_offset, y_offset, width, height)




from gi.repository import Gio


def replace_menu_section(menu: Gio.Menu, section: Gio.MenuItem):
    """Replaces an existing section in GMenu `menu` with `section`

    The sections are compared by their `id` attributes, with the
    matching section in `menu` being replaced by the passed `section`.

    If there is no section in `menu` that matches `section`'s `id`
    attribute, a ValueError is raised.
    """
    section_id = section.get_attribute_value("id").get_string()
    for idx in range(menu.get_n_items()):
        item_id = menu.get_item_attribute_value(idx, "id").get_string()
        if item_id == section_id:
            break
    else:
        # FIXME: Better exception
        raise ValueError("Section %s not found" % section_id)
    menu.remove(idx)
    menu.insert_item(idx, section)



# Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2012 Kai Willadsen <kai.willadsen@gmail.com>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from . import bzr, cvs, darcs, git, mercurial, svn

# Tuple of plugins, ordered according to best-guess as to which VC a
# user is likely to want by default in a multiple-VC situation.
VC_PLUGINS = (git, mercurial, bzr, svn, darcs, cvs)


def get_vcs(location):
    """Pick only the Vcs with the longest repo root

       Some VC plugins search their repository root
       by walking the filesystem upwards its root
       and now that we display multiple VCs in the
       same directory, we must filter those other
       repositories that are located in the search
       path towards "/" as they are not relevant
       to the user.
    """

    vcs = []
    for plugin in VC_PLUGINS:
        root, location = plugin.Vc.is_in_repo(location)
        enabled = root is not None
        vcs.append((plugin.Vc, enabled))

    return vcs



# Copyright (C) 2009 Vincent Legoll <vincent.legoll@gmail.com>
# Copyright (C) 2010-2011, 2013-2019 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from enum import Enum

from gi.repository import Gdk, Gio, GLib, GObject, Gtk, GtkSource, Pango

from meld.meldbuffer import MeldBuffer
from meld.settings import bind_settings, get_meld_settings, settings
from meld.style import colour_lookup_with_fallback, get_common_theme

log = logging.getLogger(__name__)


def get_custom_encoding_candidates():
    custom_candidates = []
    try:
        for charset in settings.get_value('detect-encodings'):
            encoding = GtkSource.Encoding.get_from_charset(charset)
            if not encoding:
                log.warning('Invalid charset "%s" skipped', charset)
                continue
            custom_candidates.append(encoding)
        if custom_candidates:
            custom_candidates.extend(
                GtkSource.Encoding.get_default_candidates())
    except AttributeError:
        # get_default_candidates() is only available in GtkSourceView 3.18
        # and we'd rather use their defaults than our old detect list.
        pass
    return custom_candidates


class LanguageManager:

    manager = GtkSource.LanguageManager()

    @classmethod
    def get_language_from_file(cls, gfile):
        try:
            info = gfile.query_info(
                Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE, 0, None)
        except (GLib.GError, AttributeError):
            return None
        content_type = info.get_content_type()
        return cls.manager.guess_language(gfile.get_basename(), content_type)

    @classmethod
    def get_language_from_mime_type(cls, mime_type):
        content_type = Gio.content_type_from_mime_type(mime_type)
        return cls.manager.guess_language(None, content_type)


class TextviewLineAnimationType(Enum):
    fill = 'fill'
    stroke = 'stroke'


class TextviewLineAnimation:
    __slots__ = ("start_mark", "end_mark", "start_rgba", "end_rgba",
                 "start_time", "duration", "anim_type")

    def __init__(self, mark0, mark1, rgba0, rgba1, duration, anim_type):
        self.start_mark = mark0
        self.end_mark = mark1
        self.start_rgba = rgba0
        self.end_rgba = rgba1
        self.start_time = GLib.get_monotonic_time()
        self.duration = duration
        self.anim_type = anim_type


class SourceViewHelperMixin:

    def get_y_for_line_num(self, line):
        buf = self.get_buffer()
        it = buf.get_iter_at_line(line)
        y, h = self.get_line_yrange(it)
        if line >= buf.get_line_count():
            return y + h
        return y

    def get_line_num_for_y(self, y):
        return self.get_line_at_y(y)[0].get_line()


class MeldSourceView(GtkSource.View, SourceViewHelperMixin):

    __gtype_name__ = "MeldSourceView"

    __gsettings_bindings_view__ = (
        ('highlight-current-line', 'highlight-current-line-local'),
        ('indent-width', 'tab-width'),
        ('insert-spaces-instead-of-tabs', 'insert-spaces-instead-of-tabs'),
        ('enable-space-drawer', 'draw-spaces-bool'),
        ('wrap-mode', 'wrap-mode'),
        ('show-line-numbers', 'show-line-numbers'),
    )

    # Named so as not to conflict with the GtkSourceView property
    highlight_current_line_local = GObject.Property(type=bool, default=False)

    def get_show_line_numbers(self):
        return self._show_line_numbers

    def set_show_line_numbers(self, show):
        if show == self._show_line_numbers:
            return

        if getattr(self, 'line_renderer', None):
            self.line_renderer.set_visible(show)

        self._show_line_numbers = bool(show)
        self.notify("show-line-numbers")

    show_line_numbers = GObject.Property(
        type=bool, default=False, getter=get_show_line_numbers,
        setter=set_show_line_numbers)

    wrap_mode_bool = GObject.Property(
        type=bool, default=False,
        nick="Wrap mode (Boolean version)",
        blurb=(
            "Mirror of the wrap-mode GtkTextView property, reduced to "
            "a single Boolean for UI ease-of-use."
        ),
    )

    draw_spaces_bool = GObject.Property(
        type=bool, default=False,
        nick="Draw spaces (Boolean version)",
        blurb=(
            "Mirror of the draw-spaces GtkSourceView property, "
            "reduced to a single Boolean for UI ease-of-use."
        ),
    )

    overscroll_num_lines = GObject.Property(
        type=int, default=5, minimum=0, maximum=100,
        nick="Overscroll line count",
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT
        ),
    )

    replaced_entries = (
        # We replace the default GtkSourceView undo mechanism
        (Gdk.KEY_z, Gdk.ModifierType.CONTROL_MASK),
        (Gdk.KEY_z, Gdk.ModifierType.CONTROL_MASK |
            Gdk.ModifierType.SHIFT_MASK),

        # We replace the default line movement behaviour of Alt+Up/Down
        (Gdk.KEY_Up, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Up, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Up, Gdk.ModifierType.MOD1_MASK |
            Gdk.ModifierType.SHIFT_MASK),
        (Gdk.KEY_Down, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Down, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Down, Gdk.ModifierType.MOD1_MASK |
            Gdk.ModifierType.SHIFT_MASK),
        # ...and Alt+Left/Right
        (Gdk.KEY_Left, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Left, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_Right, Gdk.ModifierType.MOD1_MASK),
        (Gdk.KEY_KP_Right, Gdk.ModifierType.MOD1_MASK),
        # ...and Ctrl+Page Up/Down
        (Gdk.KEY_Page_Up, Gdk.ModifierType.CONTROL_MASK),
        (Gdk.KEY_KP_Page_Up, Gdk.ModifierType.CONTROL_MASK),
        (Gdk.KEY_Page_Down, Gdk.ModifierType.CONTROL_MASK),
        (Gdk.KEY_KP_Page_Down, Gdk.ModifierType.CONTROL_MASK),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drag_dest_add_uri_targets()

        # Most bindings are on SourceView, except the Page Up/Down ones
        # which are on TextView.
        binding_set_names = ('GtkSourceView', 'GtkTextView')
        for set_name in binding_set_names:
            binding_set = Gtk.binding_set_find(set_name)
            for key, modifiers in self.replaced_entries:
                Gtk.binding_entry_remove(binding_set, key, modifiers)

        self.anim_source_id = None
        self.animating_chunks = []
        self.syncpoints = []
        self._show_line_numbers = None

        buf = MeldBuffer()
        inline_tag = GtkSource.Tag.new("inline")
        inline_tag.props.draw_spaces = True
        buf.get_tag_table().add(inline_tag)
        buf.create_tag("dimmed")
        self.set_buffer(buf)
        self.connect('notify::overscroll-num-lines', self.notify_overscroll)

    @property
    def line_height(self) -> int:
        if not getattr(self, '_approx_line_height', None):
            context = self.get_pango_context()
            layout = Pango.Layout(context)
            layout.set_text('X', -1)
            _width, self._approx_line_height = layout.get_pixel_size()

        return self._approx_line_height

    def notify_overscroll(self, view, param):
        self.props.bottom_margin = self.overscroll_num_lines * self.line_height

    def do_paste_clipboard(self, *args):
        # This is an awful hack to replace another awful hack. The idea
        # here is to sanitise the clipboard contents so that it doesn't
        # contain GtkTextTags, by requesting and setting plain text.

        def text_received_cb(clipboard, text, *user_data):
            # On clipboard failure, text will be None
            if not text:
                return

            # Manual encoding is required here, or the length will be
            # incorrect, and the API requires a UTF-8 bytestring.
            utf8_text = text.encode('utf-8')
            clipboard.set_text(text, len(utf8_text))
            self.get_buffer().paste_clipboard(
                clipboard, None, self.get_editable())

        clipboard = self.get_clipboard(Gdk.SELECTION_CLIPBOARD)
        clipboard.request_text(text_received_cb)

    def add_fading_highlight(
            self, mark0, mark1, colour_name, duration,
            anim_type=TextviewLineAnimationType.fill, starting_alpha=1.0):

        if not self.get_realized():
            return

        rgba0 = self.fill_colors[colour_name].copy()
        rgba1 = self.fill_colors[colour_name].copy()
        rgba0.alpha = starting_alpha
        rgba1.alpha = 0.0
        anim = TextviewLineAnimation(
            mark0, mark1, rgba0, rgba1, duration, anim_type)
        self.animating_chunks.append(anim)

    def on_setting_changed(self, settings, key):
        if key == 'font':
            self.override_font(settings.font)
            self._approx_line_height = None
        elif key == 'style-scheme':
            self.highlight_color = colour_lookup_with_fallback(
                "meld:current-line-highlight", "background")
            self.syncpoint_color = colour_lookup_with_fallback(
                "meld:syncpoint-outline", "foreground")
            self.fill_colors, self.line_colors = get_common_theme()

            buf = self.get_buffer()
            buf.set_style_scheme(settings.style_scheme)

            tag = buf.get_tag_table().lookup("inline")
            tag.props.background_rgba = colour_lookup_with_fallback(
                "meld:inline", "background")
            tag = buf.get_tag_table().lookup("dimmed")
            tag.props.foreground_rgba = colour_lookup_with_fallback(
                "meld:dimmed", "foreground")

    def do_realize(self):
        bind_settings(self)

        def wrap_mode_from_bool(binding, from_value):
            if from_value:
                settings_mode = settings.get_enum('wrap-mode')
                if settings_mode == Gtk.WrapMode.NONE:
                    mode = Gtk.WrapMode.WORD
                else:
                    mode = settings_mode
            else:
                mode = Gtk.WrapMode.NONE
            return mode

        def wrap_mode_to_bool(binding, from_value):
            return bool(from_value)

        self.bind_property(
            'wrap-mode-bool', self, 'wrap-mode',
            GObject.BindingFlags.BIDIRECTIONAL,
            wrap_mode_from_bool,
            wrap_mode_to_bool,
        )
        self.wrap_mode_bool = wrap_mode_to_bool(None, self.props.wrap_mode)

        self.bind_property(
            'draw-spaces-bool', self.props.space_drawer, 'enable-matrix',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
        )

        meld_settings = get_meld_settings()

        self.on_setting_changed(meld_settings, 'font')
        self.on_setting_changed(meld_settings, 'style-scheme')
        self.get_buffer().set_style_scheme(meld_settings.style_scheme)

        meld_settings.connect('changed', self.on_setting_changed)

        return GtkSource.View.do_realize(self)

    def do_unrealize(self):
        if self.anim_source_id:
            GLib.source_remove(self.anim_source_id)
        return GtkSource.View.do_unrealize(self)

    def do_draw_layer(self, layer, context):
        if layer != Gtk.TextViewLayer.BELOW_TEXT:
            return GtkSource.View.do_draw_layer(self, layer, context)

        context.save()
        context.set_line_width(1.0)

        _, clip = Gdk.cairo_get_clip_rectangle(context)
        clip_end = clip.y + clip.height
        bounds = (
            self.get_line_num_for_y(clip.y),
            self.get_line_num_for_y(clip_end),
        )

        x = clip.x - 0.5
        width = clip.width + 1

        # Paint chunk backgrounds and outlines
        for change in self.chunk_iter(bounds):
            ypos0 = self.get_y_for_line_num(change[1])
            ypos1 = self.get_y_for_line_num(change[2])
            height = max(0, ypos1 - ypos0 - 1)

            context.rectangle(x, ypos0 + 0.5, width, height)
            if change[1] != change[2]:
                context.set_source_rgba(*self.fill_colors[change[0]])
                context.fill_preserve()
                if self.current_chunk_check(change):
                    highlight = self.fill_colors['current-chunk-highlight']
                    context.set_source_rgba(*highlight)
                    context.fill_preserve()

            context.set_source_rgba(*self.line_colors[change[0]])
            context.stroke()

        textbuffer = self.get_buffer()

        # Check whether we're drawing past the last line in the buffer
        # (i.e., the overscroll) and draw a custom background if so.
        end_y, end_height = self.get_line_yrange(textbuffer.get_end_iter())
        end_y += end_height
        visible_bottom_margin = clip_end - end_y
        if visible_bottom_margin > 0:
            context.rectangle(x + 1, end_y, width - 1, visible_bottom_margin)
            context.set_source_rgba(*self.fill_colors['overscroll'])
            context.fill()

        # Paint current line highlight
        if self.props.highlight_current_line_local and self.is_focus():
            it = textbuffer.get_iter_at_mark(textbuffer.get_insert())
            ypos, line_height = self.get_line_yrange(it)
            context.rectangle(x, ypos, width, line_height)
            context.set_source_rgba(*self.highlight_color)
            context.fill()

        # Draw syncpoint indicator lines
        for syncpoint in self.syncpoints:
            if syncpoint is None:
                continue
            syncline = textbuffer.get_iter_at_mark(syncpoint).get_line()
            if bounds[0] <= syncline <= bounds[1]:
                ypos = self.get_y_for_line_num(syncline)
                context.rectangle(x, ypos - 0.5, width, 1)
                context.set_source_rgba(*self.syncpoint_color)
                context.stroke()

        # Overdraw all animated chunks, and update animation states
        new_anim_chunks = []
        for c in self.animating_chunks:
            current_time = GLib.get_monotonic_time()
            percent = min(
                1.0, (current_time - c.start_time) / float(c.duration))
            rgba_pairs = zip(c.start_rgba, c.end_rgba)
            rgba = [s + (e - s) * percent for s, e in rgba_pairs]

            it = textbuffer.get_iter_at_mark(c.start_mark)
            ystart, _ = self.get_line_yrange(it)
            it = textbuffer.get_iter_at_mark(c.end_mark)
            yend, _ = self.get_line_yrange(it)
            if ystart == yend:
                ystart -= 1

            context.set_source_rgba(*rgba)
            context.rectangle(x, ystart, width, yend - ystart)
            if c.anim_type == TextviewLineAnimationType.stroke:
                context.stroke()
            else:
                context.fill()

            if current_time <= c.start_time + c.duration:
                new_anim_chunks.append(c)
            else:
                textbuffer.delete_mark(c.start_mark)
                textbuffer.delete_mark(c.end_mark)
        self.animating_chunks = new_anim_chunks

        if self.animating_chunks and self.anim_source_id is None:
            def anim_cb():
                self.queue_draw()
                return True
            # Using timeout_add interferes with recalculation of inline
            # highlighting; this mechanism could be improved.
            self.anim_source_id = GLib.idle_add(anim_cb)
        elif not self.animating_chunks and self.anim_source_id:
            GLib.source_remove(self.anim_source_id)
            self.anim_source_id = None

        context.restore()

        return GtkSource.View.do_draw_layer(self, layer, context)


class CommitMessageSourceView(GtkSource.View):

    __gtype_name__ = "CommitMessageSourceView"

    __gsettings_bindings_view__ = (
        ('indent-width', 'tab-width'),
        ('insert-spaces-instead-of-tabs', 'insert-spaces-instead-of-tabs'),
        ('enable-space-drawer', 'enable-space-drawer'),
    )

    enable_space_drawer = GObject.Property(type=bool, default=False)

    def do_realize(self):
        bind_settings(self)

        self.bind_property(
            'enable-space-drawer', self.props.space_drawer, 'enable-matrix',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
        )

        return GtkSource.View.do_realize(self)


class MeldSourceMap(GtkSource.Map, SourceViewHelperMixin):

    __gtype_name__ = "MeldSourceMap"

    compact_view = GObject.Property(
        type=bool,
        nick="Limit the view to a fixed width",
        default=False,
    )

    COMPACT_MODE_WIDTH = 40

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connect('notify::compact-view', lambda *args: self.queue_resize())

    def do_draw_layer(self, layer, context):
        if layer != Gtk.TextViewLayer.BELOW_TEXT:
            return GtkSource.Map.do_draw_layer(self, layer, context)

        # Handle bad view assignments and partial initialisation
        parent_view = self.props.view
        if not hasattr(parent_view, 'chunk_iter'):
            return GtkSource.Map.do_draw_layer(self, layer, context)

        context.save()
        context.set_line_width(1.0)

        _, clip = Gdk.cairo_get_clip_rectangle(context)
        x = clip.x - 0.5
        width = clip.width + 1
        bounds = (
            self.get_line_num_for_y(clip.y),
            self.get_line_num_for_y(clip.y + clip.height),
        )

        # Paint chunk backgrounds
        for change in parent_view.chunk_iter(bounds):
            if change[1] == change[2]:
                # We don't have room to paint inserts in this widget
                continue

            ypos0 = self.get_y_for_line_num(change[1])
            ypos1 = self.get_y_for_line_num(change[2])
            height = max(0, ypos1 - ypos0 - 1)

            context.rectangle(x, ypos0 + 0.5, width, height)
            context.set_source_rgba(*parent_view.fill_colors[change[0]])
            context.fill()

        context.restore()

        return GtkSource.Map.do_draw_layer(self, layer, context)

    def do_get_preferred_width(self):
        if self.props.compact_view:
            return (self.COMPACT_MODE_WIDTH, self.COMPACT_MODE_WIDTH)
        else:
            return GtkSource.Map.do_get_preferred_width(self)
